from __future__ import annotations

from collections.abc import Iterable, Sequence
from functools import reduce
from itertools import chain
from math import isclose

from polars import DataFrame, Expr, PolarsDataType
from polars.exceptions import OutOfBoundsError
from polars.testing import assert_frame_equal
from polars.type_aliases import IntoExpr, JoinStrategy, JoinValidation, SchemaDict

from utilities.typing import SequenceStrs


def check_dataframe(
    df: DataFrame,
    /,
    *,
    columns: SequenceStrs | None = None,
    dtypes: list[PolarsDataType] | None = None,
    height: int | tuple[int, float] | None = None,
    min_height: int | None = None,
    max_height: int | None = None,
    schema: SchemaDict | None = None,
    shape: tuple[int, int] | None = None,
    sorted: IntoExpr | Iterable[IntoExpr] | None = None,  # noqa: A002
    unique: IntoExpr | Iterable[IntoExpr] | None = None,
    width: int | None = None,
) -> None:
    """Check the properties of a DataFrame."""
    if (columns is not None) and (df.columns != list(columns)):
        msg = f"{df=}, {columns=}"
        raise DataFrameColumnsError(msg)
    if (dtypes is not None) and (df.dtypes != dtypes):
        msg = f"{df=}, {dtypes=}"
        raise DataFrameDTypesError(msg)
    if height is not None:
        _check_dataframe_height(df, height)
    if (min_height is not None) and (len(df) < min_height):
        msg = f"{df=}, {min_height=}"
        raise DataFrameMinHeightError(msg)
    if (max_height is not None) and (len(df) > max_height):
        msg = f"{df=}, {max_height=}"
        raise DataFrameMaxHeightError(msg)
    if (schema is not None) and (df.schema != schema):
        set_act, set_exp = map(set, [df.schema, schema])
        extra = set_act - set_exp
        missing = set_exp - set_act
        differ = {
            col: (left, right)
            for col in set_act & set_exp
            if (left := df.schema[col]) != (right := schema[col])
        }
        msg = f"{df=}, {extra=}, {missing=}, {differ=}"
        raise DataFrameSchemaError(msg)
    if (shape is not None) and (df.shape != shape):
        msg = f"{df=}"
        raise DataFrameShapeError(msg)
    if sorted is not None:
        df_sorted = df.sort(sorted)
        try:
            assert_frame_equal(df, df_sorted)
        except AssertionError:
            msg = f"{df=}, {sorted=}"
            raise DataFrameSortedError(msg) from None
    if (unique is not None) and df.select(unique).is_duplicated().any():
        msg = f"{df=}, {unique=}"
        raise DataFrameUniqueError(msg)
    if (width is not None) and (df.width != width):
        msg = f"{df=}"
        raise DataFrameWidthError(msg)


class DataFrameColumnsError(Exception):
    """Raised when a DataFrame has the incorrect columns."""


class DataFrameDTypesError(Exception):
    """Raised when a DataFrame has the incorrect dtypes."""


class DataFrameMinHeightError(Exception):
    """Raised when a DataFrame does not reach the minimum height."""


class DataFrameMaxHeightError(Exception):
    """Raised when a DataFrame exceeds the maximum height."""


class DataFrameSchemaError(Exception):
    """Raised when a DataFrame has the incorrect schema."""


class DataFrameShapeError(Exception):
    """Raised when a DataFrame has the incorrect shape."""


class DataFrameSortedError(Exception):
    """Raised when a DataFrame has non-sorted values."""


class DataFrameUniqueError(Exception):
    """Raised when a DataFrame has non-unique values."""


class DataFrameWidthError(Exception):
    """Raised when a DataFrame has the incorrect width."""


def _check_dataframe_height(df: DataFrame, height: int | tuple[int, float], /) -> None:
    if isinstance(height, int) and (len(df) != height):
        msg = f"{df=}, {height=}"
        raise DataFrameHeightError(msg)
    if isinstance(height, tuple):
        height_int, rel_tol = height
        if not isclose(len(df), height_int, rel_tol=rel_tol):
            msg = f"{df=}, {height=}"
            raise DataFrameHeightError(msg)


class DataFrameHeightError(Exception):
    """Raised when a DataFrame has the incorrect height."""


def join(
    df: DataFrame,
    *dfs: DataFrame,
    on: str | Expr | Sequence[str | Expr],
    how: JoinStrategy = "inner",
    validate: JoinValidation = "m:m",
) -> DataFrame:
    def inner(left: DataFrame, right: DataFrame, /) -> DataFrame:
        return left.join(right, on=on, how=how, validate=validate)

    return reduce(inner, chain([df], dfs))


def set_first_row_as_columns(df: DataFrame, /) -> DataFrame:
    """Set the first row of a DataFrame as its columns."""

    try:
        row = df.row(0)
    except OutOfBoundsError:
        msg = f"{df=}"
        raise EmptyDataFrameError(msg) from None
    mapping = dict(zip(df.columns, row, strict=True))
    return df[1:].rename(mapping)


class EmptyDataFrameError(Exception):
    """Raised when a DataFrame is empty."""


__all__ = [
    "check_dataframe",
    "DataFrameColumnsError",
    "DataFrameDTypesError",
    "DataFrameHeightError",
    "DataFrameMaxHeightError",
    "DataFrameMinHeightError",
    "DataFrameSchemaError",
    "DataFrameShapeError",
    "DataFrameSortedError",
    "DataFrameUniqueError",
    "DataFrameWidthError",
    "EmptyDataFrameError",
    "join",
    "set_first_row_as_columns",
]



from __future__ import annotations

import datetime as dt
from collections.abc import Callable, Hashable, Iterator
from functools import partial, wraps
from inspect import signature
from pathlib import Path
from typing import Any, ParamSpec, TypeVar

from utilities.datetime import UTC, duration_to_timedelta, get_now
from utilities.git import get_repo_root_or_cwd_sub_path
from utilities.hashlib import md5_hash
from utilities.pathlib import PathLike
from utilities.pickle import read_pickle, write_pickle
from utilities.types import Duration, is_hashable

_P = ParamSpec("_P")
_R = TypeVar("_R")


def _caches(path: Path, /) -> Path:
    return path.joinpath(".caches")


_ROOT = get_repo_root_or_cwd_sub_path(_caches, if_missing=_caches)


def cache_to_disk(
    *, root: PathLike = _ROOT, ttl: Duration | None = None
) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]:
    """Factory for decorators which caches locally using pickles."""

    return partial(_cache_to_disk, root=root, ttl=ttl)


def _cache_to_disk(
    func: Callable[_P, _R], /, *, root: PathLike = _ROOT, ttl: Duration | None = None
) -> Callable[_P, _R]:
    """Decorator which caches locally using pickles."""

    root = Path(root, func.__name__)
    sig = signature(func)
    ttl_use = None if ttl is None else duration_to_timedelta(ttl)

    @wraps(func)
    def wrapped(*args: _P.args, **kwargs: _P.kwargs) -> _R:
        """The decorated function."""
        ba = sig.bind(*args, **kwargs)
        hash_args = tuple(_yield_hashable_args(*ba.args))
        hash_kwargs = tuple(_yield_hashable_kwargs(**ba.kwargs))
        path = root.joinpath(md5_hash((hash_args, hash_kwargs)))
        if _needs_run(path, ttl=ttl_use):
            value = func(*args, **kwargs)
            write_pickle(value, path, overwrite=True)
            return value
        return read_pickle(path)

    return wrapped


def _yield_hashable_args(*args: Any) -> Iterator[Hashable]:
    for i, arg in enumerate(args):
        if is_hashable(arg):
            yield arg
        else:
            msg = f"Positional argument {arg} (index {i}) is non-hashable"
            raise NonHashableArgumentError(msg)


def _yield_hashable_kwargs(**kwargs: Any) -> Iterator[tuple[str, Hashable]]:
    for key, arg in kwargs.items():
        if is_hashable(arg):
            yield key, arg
        else:
            msg = f"Keyword argument {key} = {arg} is non-hashable"
            raise NonHashableArgumentError(msg)


class NonHashableArgumentError(Exception):
    """Raised when an argument is non-hashable."""


def _needs_run(path: Path, /, *, ttl: dt.timedelta | None = None) -> bool:
    if not path.exists():
        return True
    if ttl is None:
        return False
    now = get_now()
    modified = dt.datetime.fromtimestamp(path.stat().st_mtime, tz=UTC)
    age = now - modified
    return age >= ttl



from __future__ import annotations

from collections import Counter
from collections.abc import Hashable, Iterable, Iterator, Sized
from functools import partial
from itertools import islice
from typing import Any, TypeGuard, TypeVar, cast

_T = TypeVar("_T")


def always_iterable(
    obj: _T | Iterable[_T],
    /,
    *,
    base_type: type[Any] | tuple[type[Any], ...] | None = (str, bytes),
) -> Iterator[_T]:
    """If *obj* is iterable, return an iterator over its items."""
    if obj is None:
        return iter(())
    if (base_type is not None) and isinstance(obj, base_type):
        return iter(cast(Iterable[_T], (obj,)))
    try:
        return iter(cast(Iterable[_T], obj))
    except TypeError:
        return iter(cast(Iterable[_T], (obj,)))


def check_duplicates(iterable: Iterable[Hashable], /) -> None:
    """Check if an iterable contains any duplicates."""
    dup = {k: v for k, v in Counter(iterable).items() if v > 1}
    if len(dup) >= 1:
        msg = f"{dup=}"
        raise IterableContainsDuplicatesError(msg)


class IterableContainsDuplicatesError(Exception):
    """Raised when an iterable contains duplicates."""


def chunked(
    iterable: Iterable[_T], /, *, n: int | None = None, strict: bool = False
) -> Iterator[list[_T]]:
    """Break iterable into lists of length n."""
    iterator = cast(
        Iterator[list[_T]],
        iter(partial(take, n, iter(iterable)), []),  # type: ignore
    )
    if strict:  # pragma: no cover
        if n is None:
            msg = "n must not be None when using strict mode."
            raise ValueError(msg)

        def ret() -> Iterator[list[_T]]:
            for chunk in iterator:
                if len(chunk) != n:
                    msg = "iterable is not divisible by n."
                    raise ValueError(msg)
                yield chunk

        return iter(ret())
    return iterator


def is_iterable_not_str(obj: Any, /) -> TypeGuard[Iterable[Any]]:
    """Check if an object is iterable, but not a string."""
    try:
        iter(obj)
    except TypeError:
        return False
    return not isinstance(obj, str)


def is_sized_not_str(obj: Any, /) -> TypeGuard[Sized]:
    """Check if an object is sized, but not a string."""
    try:
        _ = len(obj)
    except TypeError:
        return False
    return not isinstance(obj, str)


def one(iterable: Iterable[_T], /) -> _T:
    """Return the only item from iterable."""
    it = iter(iterable)
    try:
        first = next(it)
    except StopIteration:
        raise EmptyIterableError from None
    try:
        second = next(it)
    except StopIteration:
        return first
    else:
        msg = (
            f"Expected exactly one item in iterable, but got {first!r}, "
            f"{second!r}, and perhaps more."
        )
        raise MultipleElementsError(msg)


class EmptyIterableError(Exception):
    """Raised when an iterable is empty."""


class MultipleElementsError(Exception):
    """Raised when an iterable contains multiple elements."""


def take(n: int, iterable: Iterable[_T], /) -> list[_T]:
    """Return first n items of the iterable as a list."""
    return list(islice(iterable, n))


__all__ = [
    "always_iterable",
    "check_duplicates",
    "chunked",
    "EmptyIterableError",
    "is_iterable_not_str",
    "is_sized_not_str",
    "IterableContainsDuplicatesError",
    "MultipleElementsError",
    "one",
    "take",
]



from __future__ import annotations

from typing import Any

from pytest import mark, param

from utilities.types import (
    NoneType,
    ensure_class,
    is_hashable,
    issubclass_except_bool_int,
)


class TestEnsureClass:
    @mark.parametrize(
        ("obj", "expected"), [param(None, NoneType), param(NoneType, NoneType)]
    )
    def test_main(self, *, obj: Any, expected: type[Any]) -> None:
        assert ensure_class(obj) is expected


class TestIsHashable:
    @mark.parametrize(
        ("obj", "expected"),
        [param(0, True), param((1, 2, 3), True), param([1, 2, 3], False)],
    )
    def test_main(self, *, obj: Any, expected: bool) -> None:
        assert is_hashable(obj) is expected


class TestIsSubclassExceptBoolInt:
    @mark.parametrize(
        ("x", "y", "expected"),
        [param(bool, bool, True), param(bool, int, False), param(int, int, True)],
    )
    def test_main(self, *, x: type[Any], y: type[Any], expected: bool) -> None:
        assert issubclass_except_bool_int(x, y) is expected

    def test_subclass_of_int(self) -> None:
        class MyInt(int):
            ...

        assert not issubclass_except_bool_int(bool, MyInt)


class TestNoneType:
    def test_main(self) -> None:
        assert isinstance(None, NoneType)
